from typing import (Callable,
                    Dict,
                    Iterable,
                    List,
                    Tuple)

import pytest
from hypothesis.searchstrategy import SearchStrategy

from lz.hints import (Domain,
                      Intermediate,
                      Map,
                      Predicate,
                      Range)
from tests import strategies
from tests.utils import find


@pytest.fixture(scope='session')
def true_predicate() -> Predicate:
    return find(strategies.true_predicates)


@pytest.fixture(scope='session')
def false_predicate() -> Predicate:
    return find(strategies.false_predicates)


@pytest.fixture(scope='session')
def predicate() -> Predicate:
    return find(strategies.predicates)


@pytest.fixture(scope='session')
def left_predicate() -> Predicate:
    return find(strategies.predicates)


@pytest.fixture(scope='session')
def mid_predicate() -> Predicate:
    return find(strategies.predicates)


@pytest.fixture(scope='session')
def right_predicate() -> Predicate:
    return find(strategies.predicates)


@pytest.fixture(scope='session')
def predicate_argument() -> Domain:
    return find(strategies.predicates_arguments)


@pytest.fixture(scope='function')
def map_() -> Map:
    return find(strategies.maps)


@pytest.fixture(scope='function')
def map_argument() -> Domain:
    return find(strategies.maps_arguments)


@pytest.fixture(scope='function')
def maps() -> List[Map]:
    return find(strategies.maps_lists)


@pytest.fixture(scope='function')
def maps_arguments() -> List[Domain]:
    return find(strategies.maps_lists_arguments)


@pytest.fixture(scope='function')
def next_map(map_: Map[Domain, Range]) -> Map[Range, Intermediate]:
    return find(strategies.to_one_of_suitable_maps(map_))


@pytest.fixture(scope='function')
def last_map(next_map: Map[Range, Intermediate]) -> Map[Intermediate, Range]:
    return find(strategies.to_one_of_suitable_maps(next_map))


@pytest.fixture(scope='function')
def suitable_maps() -> Tuple[Map, ...]:
    return find(strategies.suitable_maps)


@pytest.fixture(scope='function')
def next_suitable_map(suitable_maps: Tuple[Map, ...]) -> Map:
    return find(strategies.to_one_of_suitable_maps(suitable_maps[0]))


@pytest.fixture(scope='function')
def transparent_function() -> Callable[..., Range]:
    return find(strategies.transparent_functions)


@pytest.fixture(scope='function')
def transparent_function_args(transparent_function) -> Tuple[Domain, ...]:
    return find(strategies.to_transparent_functions_args(transparent_function))


@pytest.fixture(scope='function')
def transparent_function_kwargs(transparent_function) -> Dict[str, Domain]:
    return find(strategies
                .to_transparent_functions_kwargs(transparent_function))


@pytest.fixture(scope='function')
def projector() -> Callable[[Domain, Domain], Domain]:
    return find(strategies.projectors)


@pytest.fixture(scope='function')
def projector_domain(projector: Callable[[Domain, Domain], Domain]
                     ) -> SearchStrategy:
    return find(strategies.to_projectors_domains(projector))


@pytest.fixture(scope='function')
def projector_domain_element(projector_domain: SearchStrategy) -> Domain:
    return find(projector_domain)


@pytest.fixture(scope='function')
def projector_initial(projector: Callable[[Domain, Domain], Domain],
                      projector_domain: SearchStrategy) -> Range:
    return find(strategies.to_projectors_domains_initials((projector,
                                                           projector_domain)))


@pytest.fixture(scope='function')
def projector_iterable(projector_domain: SearchStrategy) -> Iterable[Domain]:
    return find(strategies.to_iterables(projector_domain,
                                        min_size=1))



import importlib
import inspect
import os
import platform
import sys
from operator import methodcaller
from pathlib import Path
from types import (BuiltinFunctionType,
                   ModuleType)
from typing import (Any,
                    Iterable,
                    List,
                    Union)

from hypothesis import strategies
from hypothesis.searchstrategy import SearchStrategy

from lz.functional import compose
from lz.iterating import (flatmapper,
                          sifter)
from lz.logical import (conjoin,
                        negate)
from lz.signatures.hints import (MethodDescriptorType,
                                 WrapperDescriptorType)


def find_stdlib_modules_names(directory_path: Path = Path(os.__file__).parent,
                              ) -> Iterable[str]:
    yield from sys.builtin_module_names

    def is_stdlib_module_path(path: Path) -> bool:
        base_name = path.stem
        # skips 'LICENSE', '__pycache__', 'site-packages', etc.
        return not (base_name.isupper()
                    or base_name.startswith('__')
                    or '-' in base_name)

    sources_paths = filter(is_stdlib_module_path, directory_path.iterdir())
    sources_relative_paths = map(methodcaller(Path.relative_to.__name__,
                                              directory_path),
                                 sources_paths)
    yield from map(str, map(methodcaller(Path.with_suffix.__name__, ''),
                            sources_relative_paths))


stdlib_modules_names = (set(find_stdlib_modules_names())
                        - {'antigravity', 'this'})

if sys.platform == 'win32':
    stdlib_modules_names -= {'crypt', 'curses', 'pty', 'tty'}

if platform.python_implementation() == 'PyPy':
    stdlib_modules_names -= {'msilib', 'symtable', 'tracemalloc'}
    if sys.platform == 'win32':
        stdlib_modules_names -= {'asyncio'}

stdlib_modules = list(map(importlib.import_module, stdlib_modules_names))
supported_modules = set(stdlib_modules)
unsupported_modules = set()

if platform.python_implementation() != 'PyPy':
    import _collections
    import _codecs_hk
    import _codecs_iso2022
    import _codecs_jp
    import _codecs_kr
    import _codecs_cn
    import _codecs_tw
    import _lsprof
    import _multibytecodec
    import _string
    import aifc
    import audioop
    import bdb
    import ctypes
    import faulthandler
    import mailbox
    import parser
    import turtle
    import unittest
    import xxsubtype

    # not supported by ``typeshed`` package
    unsupported_modules.update({_collections,
                                _codecs_hk,
                                _codecs_iso2022,
                                _codecs_jp,
                                _codecs_kr,
                                _codecs_cn,
                                _codecs_tw,
                                _lsprof,
                                _multibytecodec,
                                _string,
                                aifc,
                                audioop,
                                bdb,
                                ctypes,
                                faulthandler,
                                mailbox,
                                parser,
                                turtle,
                                unittest,
                                xxsubtype})

    if sys.version_info >= (3, 6):
        import _sha3

        unsupported_modules.add(_sha3)

    if ((3, 6) <= sys.version_info < (3, 6, 7)
            or (3, 7) <= sys.version_info < (3, 7, 1)):
        import _blake2

        unsupported_modules.add(_blake2)

    if sys.platform == 'win32':
        import _msi

        unsupported_modules.add(_msi)

supported_modules -= unsupported_modules

is_module_supported = supported_modules.__contains__

modules = (strategies.sampled_from(stdlib_modules)
           .filter(is_module_supported))


def flatten_module_or_class(object_: Union[ModuleType, type]
                            ) -> SearchStrategy:
    return strategies.sampled_from(to_contents(object_))


def to_contents(object_: Union[ModuleType, type]) -> List[Any]:
    return list(vars(object_).values())


to_callables = compose(sifter(callable),
                       to_contents)
unsupported_modules_callables = list(flatmapper(to_callables)
                                     (unsupported_modules))

is_callable_supported = negate(unsupported_modules_callables.__contains__)

modules_callables = (modules.flatmap(flatten_module_or_class)
                     .filter(callable)
                     .filter(is_callable_supported))


def is_not_private(object_: Union[BuiltinFunctionType,
                                  MethodDescriptorType,
                                  type]) -> bool:
    return not object_.__name__.startswith('_')


unsupported_classes = set()

if platform.python_implementation() != 'PyPy':
    import _ast
    import _collections_abc
    import _io
    import _json
    import _ssl
    import _thread
    import _weakrefset
    import asyncio.events
    import encodings
    import imaplib
    import itertools
    import macpath
    import pdb
    import random
    import runpy
    import smtplib
    import socket
    import tarfile
    import tkinter
    import types
    import warnings

    # not supported by ``typeshed`` package
    unsupported_classes.update({_ast.excepthandler,
                                _collections_abc.mappingproxy,
                                _io._BufferedIOBase,
                                _io._IOBase,
                                _io._RawIOBase,
                                _io._TextIOBase,
                                _ssl._SSLContext,
                                _thread.RLock,
                                _thread._local,
                                _weakrefset.ref,
                                asyncio.events._RunningLoop,
                                imaplib.IMAP4.abort,
                                imaplib.IMAP4.error,
                                imaplib.IMAP4.readonly,
                                itertools._grouper,
                                itertools._tee,
                                itertools._tee_dataobject,
                                encodings.CodecRegistryError,
                                macpath.norm_error,
                                pdb._rstr,
                                pdb.Restart,
                                random._MethodType,
                                runpy._Error,
                                smtplib.SMTPNotSupportedError,
                                socket._GiveupOnSendfile,
                                tarfile.EmptyHeaderError,
                                tarfile.EOFHeaderError,
                                tarfile.InvalidHeaderError,
                                tarfile.TruncatedHeaderError,
                                tarfile.SubsequentHeaderError,
                                tkinter.TclError,
                                warnings._OptionError})

    if sys.version_info >= (3, 6):
        import asyncio.base_futures

        unsupported_classes.update({_ast.Constant,
                                    asyncio.base_futures.InvalidStateError})

    if sys.version_info < (3, 7):
        import plistlib

        unsupported_classes.update({_collections_abc.range_iterator,
                                    os.terminal_size,
                                    os.times_result,
                                    os.uname_result,
                                    plistlib._InternalDict})
    else:
        import dataclasses
        import typing

        unsupported_classes.update({asyncio.events.SendfileNotAvailableError,
                                    dataclasses._InitVarMeta,
                                    typing._ProtocolMeta})

    if sys.platform == 'win32':
        import msilib
        import subprocess

        unsupported_classes.update({msilib.UuidCreate,
                                    msilib.FCICreate,
                                    msilib.OpenDatabase,
                                    msilib.CreateRecord,
                                    msilib.MSIError,
                                    subprocess.Handle})

        if sys.version_info < (3, 7):
            unsupported_classes.add(os.statvfs_result)
    else:
        import pwd
        import termios

        unsupported_classes.update({os.waitid_result,
                                    pwd.struct_passwd,
                                    termios.error})

is_class_supported = conjoin(is_not_private,
                             negate(unsupported_classes.__contains__))

classes = (modules_callables.filter(inspect.isclass)
           .filter(is_class_supported))
classes_callables = (classes.flatmap(flatten_module_or_class)
                     .filter(callable)
                     .filter(is_callable_supported))
methods = classes_callables.filter(inspect.isfunction)


def is_method_descriptor(object_: Any) -> bool:
    return isinstance(object_, MethodDescriptorType)


unsupported_methods_descriptors = set()

if platform.python_implementation() != 'PyPy':
    import _collections_abc
    import _io
    import _thread
    import collections
    import sqlite3
    import types

    # not supported by ``typeshed`` package
    unsupported_methods_descriptors.update(
            {_collections_abc.dict_items.isdisjoint,
             _collections_abc.generator.close,
             _collections_abc.generator.send,
             _collections_abc.generator.throw,
             _collections_abc.coroutine.close,
             _collections_abc.coroutine.send,
             _collections_abc.coroutine.throw,
             _io.BufferedRWPair.peek,
             _thread.LockType.acquire,
             _thread.LockType.acquire_lock,
             _thread.LockType.locked,
             _thread.LockType.locked_lock,
             _thread.LockType.release,
             _thread.LockType.release_lock,
             collections.OrderedDict.clear,
             collections.OrderedDict.pop,
             collections.OrderedDict.update,
             int.conjugate,
             sqlite3.Connection.enable_load_extension,
             sqlite3.Connection.load_extension,
             types.FrameType.clear})

    if sys.version_info >= (3, 6):
        unsupported_methods_descriptors.update(
                {_collections_abc.async_generator.aclose,
                 _collections_abc.async_generator.asend,
                 _collections_abc.async_generator.athrow})
        if sys.platform == 'linux':
            import socket

            unsupported_methods_descriptors.add(socket.socket.sendmsg_afalg)

    if sys.version_info >= (3, 7):
        import socket

        unsupported_methods_descriptors.update({bytes.isascii,
                                                bytearray.isascii,
                                                socket.socket.getblocking})
    else:
        import zipimport

        unsupported_methods_descriptors.update(
                {collections.OrderedDict.setdefault,
                 dict.get,
                 float.conjugate,
                 zipimport.zipimporter.find_loader})

    if sys.platform == 'win32':
        import socket

        unsupported_methods_descriptors.add(socket.socket.share)

is_method_descriptor_supported = conjoin(
        is_not_private,
        negate(unsupported_methods_descriptors.__contains__))

methods_descriptors = (classes_callables.filter(is_method_descriptor)
                       .filter(is_method_descriptor_supported))


def is_wrapper_descriptor(object_: Any) -> bool:
    return isinstance(object_, WrapperDescriptorType)


unsupported_wrappers_descriptors = set()

if platform.python_implementation() != 'PyPy':
    import _collections_abc

    # not supported by ``typeshed`` package
    unsupported_wrappers_descriptors.update({
        _collections_abc.coroutine.__del__,
        _collections_abc.generator.__del__})

    if sys.version_info >= (3, 6):
        import _socket

        unsupported_wrappers_descriptors.add(_socket.socket.__del__)

is_wrapper_descriptor_supported = negate(unsupported_wrappers_descriptors
                                         .__contains__)

wrappers_descriptors = (classes_callables.filter(is_wrapper_descriptor)
                        .filter(is_wrapper_descriptor_supported))
functions = modules_callables.filter(inspect.isfunction)

unsupported_built_in_functions = set()

if platform.python_implementation() != 'PyPy':
    import _hashlib
    import _json
    import _thread
    import codecs
    import socket

    # not supported by ``typeshed`` package
    unsupported_built_in_functions.update({_hashlib.openssl_md5,
                                           _hashlib.openssl_sha1,
                                           _hashlib.openssl_sha224,
                                           _hashlib.openssl_sha256,
                                           _hashlib.openssl_sha384,
                                           _hashlib.openssl_sha512,
                                           _json.encode_basestring,
                                           _thread.allocate,
                                           _thread.exit_thread,
                                           _thread.interrupt_main,
                                           _thread.stack_size,
                                           _thread.start_new_thread,
                                           codecs.backslashreplace_errors,
                                           codecs.ignore_errors,
                                           codecs.namereplace_errors,
                                           codecs.replace_errors,
                                           codecs.strict_errors,
                                           codecs.xmlcharrefreplace_errors,
                                           socket.dup,
                                           sys.callstats,
                                           sys.getallocatedblocks,
                                           sys.get_coroutine_wrapper,
                                           sys.set_coroutine_wrapper})
    if sys.version_info >= (3, 6):
        unsupported_built_in_functions.update({sys.getfilesystemencodeerrors,
                                               sys.get_asyncgen_hooks,
                                               sys.set_asyncgen_hooks})

    if sys.version_info >= (3, 7):
        unsupported_built_in_functions.update({socket.close,
                                               sys.breakpointhook})

    if sys.platform != 'win32':
        import _locale

        unsupported_built_in_functions.update(
                {_locale.bind_textdomain_codeset,
                 _locale.bindtextdomain,
                 _locale.dcgettext,
                 _locale.dgettext,
                 _locale.gettext,
                 _locale.textdomain})

        if sys.version_info >= (3, 7):
            import time

            unsupported_built_in_functions.add(time.pthread_getcpuclockid)


def has_module(function: BuiltinFunctionType) -> bool:
    return bool(function.__module__)


is_built_in_function_supported = conjoin(
        is_not_private,
        has_module,
        negate(unsupported_built_in_functions.__contains__))

built_in_functions = (modules_callables.filter(inspect.isbuiltin)
                      .filter(is_built_in_function_supported))
unsupported_callables = strategies.sampled_from(
        list(unsupported_built_in_functions
             | unsupported_classes
             | unsupported_methods_descriptors
             | unsupported_wrappers_descriptors))
callables = (built_in_functions
             | classes
             | functions
             | methods
             | methods_descriptors
             | wrappers_descriptors)



import platform

from .base import (Base,
                   Parameter,
                   Plain,
                   factory)

if platform.python_implementation() != 'PyPy':
    from .base import Overloaded

del platform



from functools import wraps
from typing import (Any,
                    Hashable,
                    MutableMapping)

from lz.hints import (Domain,
                      Map,
                      Operator,
                      Range)


def cached_map(cache: MutableMapping[Hashable, Any],
               *,
               update: bool = False) -> Operator[Map[Domain, Range]]:
    def wrapper(map_: Map[Domain, Range]) -> Map[Domain, Range]:
        @wraps(map_)
        def wrapped(argument: Domain) -> Range:
            try:
                return cache[argument]
            except KeyError:
                result = map_(argument)
                if update:
                    cache[argument] = result
                return result

        return wrapped

    return wrapper



import os
from itertools import repeat
from operator import (itemgetter,
                      truediv)
from pathlib import Path
from typing import (Iterator,
                    List)

from lz.functional import (compose,
                           pack)
from lz.iterating import (flatmapper,
                          mapper)

INIT_MODULE_NAME = '__init__'


def find_files(directory: Path) -> Iterator[Path]:
    def to_paths(root: str, files: List[str]) -> Iterator[Path]:
        yield from map(truediv, repeat(Path(root)), files)

    finder = compose(flatmapper(pack(to_paths)),
                     mapper(itemgetter(0, 2)),
                     os.walk,
                     str)
    yield from finder(directory)



import builtins
import importlib
from contextlib import suppress
from functools import singledispatch
from itertools import chain
from pathlib import Path
from types import MappingProxyType
from typing import (Any,
                    Dict,
                    Iterable,
                    List)

from typed_ast import ast3

from lz import (left,
                right)
from lz.iterating import expand
from . import (catalog,
               dictionaries,
               namespaces,
               sources)
from .conversion import TypedToPlain
from .hints import Namespace

Nodes = Dict[catalog.Path, ast3.AST]

TYPING_MODULE_PATH = catalog.factory('typing')
OVERLOAD_DECORATORS_PATHS = {catalog.factory('typing.overload'),
                             catalog.factory('overload')}


def to_nodes(object_path: catalog.Path,
             module_path: catalog.Path) -> List[ast3.AST]:
    nodes = module_path_to_nodes(module_path,
                                 base=built_ins_nodes)
    reduce_node = Reducer(nodes=nodes,
                          parent_path=catalog.Path()).visit
    reduce_node(nodes[catalog.Path()])
    try:
        result = nodes[object_path]
    except KeyError:
        reduce_node(nodes[object_path.parent])
        result = nodes[object_path]
    if not isinstance(result, list):
        result = [result]
    return result


def module_path_to_nodes(module_path: catalog.Path,
                         *,
                         base: Nodes = MappingProxyType({})) -> Nodes:
    source_path = sources.factory(module_path)
    module_root = factory(source_path)
    namespace = namespaces.factory(module_path)
    namespace = dictionaries.merge([built_ins_namespace, namespace])
    Flattener(namespace=namespace,
              module_path=module_path,
              parent_path=catalog.Path()).visit(module_root)
    result = dict(base)
    Registry(nodes=result,
             parent_path=catalog.Path()).visit(module_root)
    return result


@singledispatch
def factory(object_: Any) -> ast3.Module:
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(object_)))


@factory.register(Path)
def from_source_path(path: Path) -> ast3.Module:
    return ast3.parse(path.read_text())


@factory.register(catalog.Path)
def from_module_path(path: catalog.Path) -> ast3.Module:
    return factory(sources.factory(path))


@factory.register(ast3.Module)
def from_root_node(node: ast3.Module) -> ast3.Module:
    return node


@factory.register(ast3.AST)
def from_node(node: ast3.AST) -> ast3.Module:
    return ast3.Module([node], [])


class Base(ast3.NodeTransformer):
    def __init__(self,
                 *,
                 parent_path: catalog.Path,
                 is_nested: bool = False) -> None:
        self.parent_path = parent_path
        self.is_nested = is_nested

    def resolve_path(self, path: catalog.Path) -> catalog.Path:
        if self.is_nested:
            return self.parent_path.join(path)
        return path


class Flattener(Base):
    def __init__(self,
                 *,
                 namespace: Namespace,
                 module_path: catalog.Path,
                 parent_path: catalog.Path,
                 is_nested: bool = False) -> None:
        super().__init__(parent_path=parent_path,
                         is_nested=is_nested)
        self.namespace = namespace
        self.module_path = module_path

    def visit_Import(self, node: ast3.Import) -> Iterable[ast3.Import]:
        for name_alias in node.names:
            alias_path = self.resolve_path(to_alias_path(name_alias))
            actual_path = to_actual_path(name_alias)
            if not namespace_contains(self.namespace, alias_path):
                parent_module_name = actual_path.parts[0]
                module = importlib.import_module(parent_module_name)
                self.namespace[parent_module_name] = module
            yield ast3.Import([name_alias])

    def visit_ImportFrom(self, node: ast3.ImportFrom
                         ) -> Iterable[ast3.ImportFrom]:
        parent_module_path = to_parent_module_path(
                node,
                parent_module_path=self.module_path)
        for name_alias in node.names:
            alias_path = self.resolve_path(to_alias_path(name_alias))
            actual_path = to_actual_path(name_alias)
            if actual_path == catalog.WILDCARD_IMPORT:
                self.namespace.update(namespaces.factory(parent_module_path))
                yield from module_path_to_nodes(parent_module_path).values()
                continue
            elif not namespace_contains(self.namespace, alias_path):
                namespace = namespaces.factory(parent_module_path)
                try:
                    self.namespace[str(alias_path)] = search_by_path(
                            namespace,
                            actual_path)
                except KeyError:
                    module_path = parent_module_path.join(actual_path)
                    with suppress(ImportError):
                        self.namespace[str(alias_path)] = (
                            importlib.import_module(str(module_path)))
            yield ast3.ImportFrom(str(parent_module_path), [name_alias], 0)

    def visit_ClassDef(self, node: ast3.ClassDef) -> ast3.ClassDef:
        path = self.resolve_path(catalog.factory(node.name))
        transformer = type(self)(namespace=self.namespace,
                                 parent_path=path,
                                 module_path=self.module_path,
                                 is_nested=True)
        for child in node.body:
            transformer.visit(child)
        return node

    def visit_If(self, node: ast3.If) -> Iterable[ast3.AST]:
        if self.visit(node.test):
            children = node.body
        else:
            children = node.orelse
        for child in children:
            self.visit(child)
        yield from children

    def visit_BoolOp(self, node: ast3.BoolOp) -> bool:
        return self.evaluate_expression(node)

    def visit_Compare(self, node: ast3.Compare) -> bool:
        return self.evaluate_expression(node)

    def evaluate_expression(self, node: ast3.expr) -> Any:
        # to avoid name conflicts
        # we're using name that won't be present
        # because it'll lead to ``SyntaxError`` otherwise
        # and no AST will be generated
        temporary_name = '@tmp'
        assignment = expression_to_assignment(node,
                                              name=temporary_name)
        execute(assignment,
                namespace=self.namespace)
        return self.namespace.pop(temporary_name)


class Registry(Base):
    def __init__(self,
                 *,
                 nodes: Nodes,
                 parent_path: catalog.Path,
                 is_nested: bool = False) -> None:
        super().__init__(parent_path=parent_path,
                         is_nested=is_nested)
        self.nodes = nodes

    def visit_Module(self, node: ast3.Module) -> ast3.Module:
        self.nodes[catalog.Path()] = node
        for child in node.body:
            self.visit(child)
        return node

    def visit_ClassDef(self, node: ast3.ClassDef) -> ast3.ClassDef:
        path = self.resolve_path(catalog.factory(node.name))
        try:
            self.nodes[path]
        except KeyError:
            pass
        else:
            inherits_itself = any(path == base_path
                                  for base_path in map(self.visit, node.bases))
            if not inherits_itself:
                nodes = {object_path: object_node
                         for object_path, object_node in self.nodes.items()
                         if not object_path.is_child_of(path)}
                self.nodes.clear()
                self.nodes.update(nodes)
        self.nodes[path] = node
        transformer = type(self)(nodes=self.nodes,
                                 parent_path=path,
                                 is_nested=True)
        for child in node.body:
            transformer.visit(child)
        return node

    def visit_FunctionDef(self, node: ast3.FunctionDef) -> ast3.FunctionDef:
        path = self.resolve_path(catalog.factory(node.name))
        if self.is_overloaded(node):
            self.nodes.setdefault(path, []).append(node)
        else:
            self.nodes[path] = node
        return node

    def is_overloaded(self, node: ast3.FunctionDef) -> bool:
        decorators_paths = map(self.visit, node.decorator_list)
        overload_decorators_paths = (OVERLOAD_DECORATORS_PATHS
                                     & set(decorators_paths))
        try:
            overload_decorator_path, = overload_decorators_paths
        except ValueError:
            pass
        else:
            root_path = catalog.factory(overload_decorator_path.parts[0])
            root_node = self.nodes[root_path]
            if isinstance(root_node, ast3.Import):
                imported_modules_paths = map(to_actual_path, root_node.names)
                if any(module_path == TYPING_MODULE_PATH
                       for module_path in imported_modules_paths):
                    return True
            elif isinstance(root_node, ast3.ImportFrom):
                imported_from_module_path = catalog.factory(root_node.module)
                if imported_from_module_path == TYPING_MODULE_PATH:
                    return True
        return False

    def visit_Import(self, node: ast3.Import) -> ast3.Import:
        for child in node.names:
            alias_path = self.resolve_path(to_alias_path(child))
            self.nodes[alias_path] = node
        return node

    def visit_ImportFrom(self, node: ast3.ImportFrom) -> ast3.ImportFrom:
        for child in node.names:
            alias_path = self.resolve_path(to_alias_path(child))
            self.nodes[alias_path] = node
        return node

    def visit_Assign(self, node: ast3.Assign) -> ast3.Assign:
        paths = map(self.visit, node.targets)
        value_node = node.value
        for path in paths:
            self.nodes[path] = value_node
        return node

    def visit_AnnAssign(self, node: ast3.AnnAssign) -> ast3.AnnAssign:
        path = self.visit(node.target)
        self.nodes[path] = node
        return node

    def visit_Name(self, node: ast3.Name) -> catalog.Path:
        name_path = catalog.factory(node.id)
        if isinstance(node.ctx, ast3.Load):
            return name_path
        elif isinstance(node.ctx, ast3.Store):
            return self.resolve_path(name_path)

    def visit_Attribute(self, node: ast3.Attribute) -> catalog.Path:
        parent_path = self.visit(node.value)
        attribute_path = catalog.factory(node.attr)
        return parent_path.join(attribute_path)

    def visit_Subscript(self, node: ast3.Subscript) -> catalog.Path:
        context = node.ctx
        if not isinstance(context, ast3.Load):
            raise TypeError('Unsupported context type: {type}.'
                            .format(type=type(context)))
        return self.visit(node.value)


complete_new_style_class_bases = right.attacher(ast3.Name('object',
                                                          ast3.Load()))


class Reducer(Base):
    def __init__(self,
                 *,
                 nodes: Nodes,
                 parent_path: catalog.Path,
                 is_nested: bool = False) -> None:
        super().__init__(parent_path=parent_path,
                         is_nested=is_nested)
        self.nodes = nodes

    def visit_Import(self, node: ast3.Import) -> ast3.Import:
        for child in node.names:
            alias_path = to_alias_path(child)
            actual_path = to_actual_path(child)
            nodes = module_path_to_nodes(actual_path)
            self.nodes.update(dict(zip(map(alias_path.join, nodes.keys()),
                                       nodes.values())))
        return node

    def visit_ImportFrom(self, node: ast3.ImportFrom) -> ast3.ImportFrom:
        parent_module_path = catalog.factory(node.module)
        for name_alias in node.names:
            alias_path = self.resolve_path(to_alias_path(name_alias))
            actual_path = to_actual_path(name_alias)
            if actual_path == catalog.WILDCARD_IMPORT:
                nodes = module_path_to_nodes(parent_module_path)
                self.nodes.update(nodes)
                continue
            object_path = parent_module_path.join(actual_path)
            if is_module_path(object_path):
                module_root = factory(object_path)
                self.nodes[alias_path] = module_root
            else:
                nodes = module_path_to_nodes(parent_module_path)
                target_node = nodes.pop(actual_path)
                if isinstance(target_node, (ast3.Import, ast3.ImportFrom)):
                    # handle chained imports
                    nodes = {}
                    transformer = type(self)(nodes=nodes,
                                             parent_path=parent_module_path)
                    transformer.visit(target_node)
                    target_node = nodes.pop(actual_path)
                self.nodes[alias_path] = target_node
                # hack to be able to visit "imported" nodes
                self.nodes.update(nodes)
        return node

    def visit_ClassDef(self, node: ast3.ClassDef) -> ast3.ClassDef:
        path = self.resolve_path(catalog.factory(node.name))
        bases = complete_new_style_class_bases(node.bases)
        for base_path in map(self.visit, bases):
            base_nodes = {object_path: node
                          for object_path, node in self.nodes.items()
                          if object_path.is_child_of(base_path)
                          and object_path != base_path}
            for base_object_path, base_object_node in base_nodes.items():
                self.nodes.setdefault(base_object_path.with_parent(path),
                                      base_object_node)
        transformer = type(self)(nodes=self.nodes,
                                 parent_path=path,
                                 is_nested=True)
        for child in node.body:
            transformer.visit(child)
        return node

    def visit_FunctionDef(self, node: ast3.FunctionDef) -> ast3.FunctionDef:
        return node

    def visit_Assign(self, node: ast3.Assign) -> ast3.Assign:
        return node

    def visit_AnnAssign(self, node: ast3.AnnAssign) -> ast3.AnnAssign:
        return node

    def visit_Call(self, node: ast3.Call) -> catalog.Path:
        return self.visit(node.func)

    def visit_Attribute(self, node: ast3.Attribute) -> catalog.Path:
        parent_path = self.visit(node.value)
        attribute_path = catalog.factory(node.attr)
        object_path = parent_path.join(attribute_path)
        self.nodes[attribute_path] = self.nodes[object_path]
        return object_path

    def visit_Name(self, node: ast3.Name) -> catalog.Path:
        return self.resolve_path(catalog.factory(node.id))

    def visit_Subscript(self, node: ast3.Subscript) -> catalog.Path:
        context = node.ctx
        if not isinstance(context, ast3.Load):
            raise TypeError('Unsupported context type: {type}.'
                            .format(type=type(context)))
        return self.visit(node.value)


def search_by_path(namespace: Namespace, path: catalog.Path) -> Any:
    return left.folder(getattr, namespace[path.parts[0]])(path.parts[1:])


def namespace_contains(namespace: Namespace, path: catalog.Path) -> bool:
    try:
        search_by_path(namespace, path)
    except (KeyError, AttributeError):
        return False
    else:
        return True


def is_module_path(object_path: catalog.Path) -> bool:
    try:
        importlib.import_module(str(object_path))
    except ImportError:
        return False
    else:
        return True


def to_parent_module_path(object_: ast3.ImportFrom,
                          *,
                          parent_module_path: catalog.Path) -> catalog.Path:
    level = object_.level
    import_is_relative = level > 0
    if not import_is_relative:
        return catalog.factory(object_.module)
    depth = (len(parent_module_path.parts)
             + is_package(parent_module_path)
             - level) or None
    module_path_parts = filter(None,
                               chain(parent_module_path.parts[:depth],
                                     expand(object_.module)))
    return catalog.Path(*module_path_parts)


def is_package(module_path: catalog.Path) -> bool:
    return hasattr(importlib.import_module(str(module_path)), '__path__')


def to_alias_path(node: ast3.alias) -> catalog.Path:
    result = node.asname
    if result is None:
        result = node.name
    return catalog.factory(result)


def to_actual_path(node: ast3.alias) -> catalog.Path:
    return catalog.factory(node.name)


def expression_to_assignment(node: ast3.expr,
                             *,
                             name: str) -> ast3.Assign:
    name_node = ast3.Name(name, ast3.Store())
    result = ast3.Assign([name_node], node, None)
    return ast3.fix_missing_locations(result)


@singledispatch
def execute(node: ast3.AST,
            *,
            namespace: Namespace) -> None:
    raise TypeError('Unsupported node type: {type}.'
                    .format(type=type(node)))


@execute.register(ast3.stmt)
def execute_statement(node: ast3.stmt,
                      *,
                      namespace: Namespace) -> None:
    execute_tree(factory(node),
                 namespace=namespace)


@execute.register(ast3.Module)
def execute_tree(node: ast3.Module,
                 *,
                 namespace: Namespace) -> None:
    node = TypedToPlain().visit(node)
    code = compile(node, '<unknown>', 'exec')
    exec(code, namespace)


built_ins_namespace = namespaces.factory(builtins)
built_ins_nodes = module_path_to_nodes(catalog.factory(builtins))



from typing import (Any,
                    Dict)

Namespace = Dict[str, Any]

try:
    from types import MethodDescriptorType
except ImportError:
    MethodDescriptorType = type(list.append)
try:
    from types import WrapperDescriptorType
except ImportError:
    WrapperDescriptorType = type(list.__init__)



import enum
import inspect
import platform
from abc import (ABC,
                 abstractmethod)
from collections import defaultdict
from functools import (lru_cache,
                       partial,
                       singledispatch,
                       wraps)
from itertools import (product,
                       starmap,
                       zip_longest)
from operator import (attrgetter,
                      methodcaller)
from types import (BuiltinFunctionType,
                   BuiltinMethodType,
                   FunctionType,
                   MethodType)
from typing import (Callable,
                    Dict,
                    Iterable,
                    List,
                    Optional)

from lz import right
from lz.functional import (cleave,
                           compose,
                           identity,
                           pack)
from lz.hints import (Domain,
                      Map,
                      Range)
from lz.iterating import (cutter,
                          expand,
                          first,
                          flatten,
                          grouper,
                          mapper,
                          reverse,
                          sifter)
from .hints import (MethodDescriptorType,
                    WrapperDescriptorType)
from .utils import cached_map


class Parameter:
    class Kind(enum.IntEnum):
        POSITIONAL_ONLY = 0
        POSITIONAL_OR_KEYWORD = 1
        VARIADIC_POSITIONAL = 2
        KEYWORD_ONLY = 3
        VARIADIC_KEYWORD = 4

        def __repr__(self) -> str:
            return type(self).__qualname__ + '.' + self._name_

    kinds_prefixes = defaultdict(str,
                                 {Kind.VARIADIC_POSITIONAL: '*',
                                  Kind.VARIADIC_KEYWORD: '**'})

    def __init__(self,
                 *,
                 name: str,
                 kind: Kind,
                 has_default: bool) -> None:
        self.name = name
        self.kind = kind
        self.has_default = has_default

    def __eq__(self, other: 'Parameter') -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        return (self.name == other.name
                and self.kind == other.kind
                and self.has_default is other.has_default)

    def __hash__(self) -> int:
        return hash((self.name, self.kind, self.has_default))

    def __repr__(self) -> str:
        return (type(self).__qualname__
                + '('
                + ', '.join(starmap('{}={!r}'.format, vars(self).items()))
                + ')')

    def __str__(self) -> str:
        return ''.join([self.kinds_prefixes[self.kind],
                        self.name,
                        '=...' if self.has_default else ''])


to_parameters_by_kind = compose(partial(defaultdict, list),
                                grouper(attrgetter('kind')))
to_parameters_by_name = compose(dict,
                                mapper(compose(tuple,
                                               cleave([attrgetter('name'),
                                                       identity]))))
all_parameters_has_defaults = compose(all,
                                      mapper(attrgetter('has_default')))


class Base(ABC):
    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __eq__(self, other: 'Base') -> bool:
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def has_unset_parameters(self, *args: Domain, **kwargs: Domain) -> bool:
        pass


class Plain(Base):
    def __init__(self, *parameters: Parameter) -> None:
        self.parameters = parameters

    def __eq__(self, other: Base) -> bool:
        if not isinstance(other, Base):
            return NotImplemented
        if not isinstance(other, Plain):
            return False
        return self.parameters == other.parameters

    def __hash__(self) -> int:
        return hash(self.parameters)

    def __repr__(self) -> str:
        return (type(self).__qualname__
                + '(' + ', '.join(map(repr, self.parameters)) + ')')

    def __str__(self) -> str:
        return '(' + ', '.join(map(str, self.parameters)) + ')'

    @property
    @lru_cache(None)
    def parameters_by_kind(self) -> Dict[Parameter.Kind, List[Parameter]]:
        return to_parameters_by_kind(self.parameters)

    def has_unset_parameters(self, *args: Domain, **kwargs: Domain) -> bool:
        positionals = (
                self.parameters_by_kind[Parameter.Kind.POSITIONAL_ONLY]
                + self.parameters_by_kind[
                    Parameter.Kind.POSITIONAL_OR_KEYWORD])
        unexpected_positional_arguments_found = (
                not self.parameters_by_kind[Parameter.Kind.VARIADIC_POSITIONAL]
                and len(args) > len(positionals))
        if unexpected_positional_arguments_found:
            return False
        to_rest_positionals = cutter(slice(len(args), None))
        rest_positionals = to_rest_positionals(positionals)
        rest_positionals_by_kind = to_parameters_by_kind(rest_positionals)
        rest_keywords = (
                rest_positionals_by_kind[Parameter.Kind.POSITIONAL_OR_KEYWORD]
                + self.parameters_by_kind[Parameter.Kind.KEYWORD_ONLY])
        rest_keywords_by_name = to_parameters_by_name(rest_keywords)
        unexpected_keyword_arguments_found = (
                not self.parameters_by_kind[Parameter.Kind.VARIADIC_KEYWORD]
                and kwargs.keys() - rest_keywords_by_name.keys())
        if unexpected_keyword_arguments_found:
            return False
        rest_keywords_by_name = {
            name: parameter
            for name, parameter in rest_keywords_by_name.items()
            if name not in kwargs}
        rest_positionals_only = (
            rest_positionals_by_kind[Parameter.Kind.POSITIONAL_ONLY])
        rest_keywords = rest_keywords_by_name.values()
        return not (all_parameters_has_defaults(rest_positionals_only)
                    and all_parameters_has_defaults(rest_keywords))


@singledispatch
def factory(object_: Callable[..., Range]) -> Base:
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(object_)))


def from_callable(object_: Callable[..., Range]) -> Base:
    raw_signature = inspect.signature(object_)

    def normalize_parameter(raw_parameter: inspect.Parameter) -> Parameter:
        has_default = raw_parameter.default is not inspect._empty
        return Parameter(name=raw_parameter.name,
                         kind=Parameter.Kind(raw_parameter.kind),
                         has_default=has_default)

    parameters = map(normalize_parameter, raw_signature.parameters.values())
    return Plain(*parameters)


@singledispatch
def slice_parameters(signature: Base,
                     slice_: slice) -> Base:
    raise TypeError('Unsupported signature type: {type}.'
                    .format(type=type(signature)))


@slice_parameters.register(Plain)
def slice_plain_parameters(signature: Plain,
                           slice_: slice) -> Base:
    return Plain(*signature.parameters[slice_])


if platform.python_implementation() == 'PyPy':
    from lz.iterating import slider


    def from_class(object_: type) -> Base:
        try:
            return from_callable(object_)
        except ValueError:
            method = None
            for cls, next_cls in slider(2)(object_.__mro__):
                if cls.__new__ is not next_cls.__new__:
                    method = cls.__new__
                    break
                elif cls.__init__ is not next_cls.__init__:
                    method = cls.__init__
                    break
            if method is None:
                raise
            return slice_parameters(factory(method), slice(1, None))
else:
    from typed_ast import ast3

    from . import (arboretum,
                   catalog)


    class Overloaded(Base):
        def __init__(self, *signatures: Base) -> None:
            self.signatures = frozenset(signatures)

        def __eq__(self, other: Base) -> bool:
            if not isinstance(other, Base):
                return NotImplemented
            if not isinstance(other, Overloaded):
                return False
            return self.signatures == other.signatures

        def __hash__(self) -> int:
            return hash(self.signatures)

        def __repr__(self) -> str:
            return (type(self).__qualname__
                    + '(' + ', '.join(map(repr, self.signatures)) + ')')

        def __str__(self) -> str:
            return '\nor\n'.join(map(str, self.signatures))

        def has_unset_parameters(self, *args: Domain, **kwargs: Domain
                                 ) -> bool:
            return all(map(methodcaller(Base.has_unset_parameters.__name__,
                                        *args, **kwargs),
                           self.signatures))


    def with_typeshed(function: Map[Callable[..., Range], Base]
                      ) -> Map[Callable[..., Range], Base]:
        @wraps(function)
        def wrapped(object_: Callable[..., Range]) -> Base:
            try:
                return function(object_)
            except ValueError as error:
                object_paths = catalog.paths_factory(object_)
                module_path = catalog.factory(catalog
                                              .module_name_factory(object_))
                signature_factory = partial(to_signature,
                                            module_path=module_path)
                signatures_factory = compose(sifter(),
                                             mapper(signature_factory))
                try:
                    return first(signatures_factory(object_paths))
                except StopIteration:
                    raise error

        return wrapped


    from_callable = with_typeshed(from_callable)

    to_methods_paths = mapper(methodcaller(catalog.Path.join.__name__,
                                           catalog.factory('__init__')))


    def from_class(object_: type) -> Base:
        try:
            return from_callable(object_)
        except ValueError as error:
            method_paths = to_methods_paths(catalog.paths_factory(object_))
            module_path = catalog.factory(catalog
                                          .module_name_factory(object_))
            signature_factory = partial(to_signature,
                                        module_path=module_path)
            signatures_factory = compose(sifter(),
                                         mapper(signature_factory))
            try:
                method_signature = first(signatures_factory(method_paths))
            except StopIteration:
                raise error
            else:
                return slice_parameters(method_signature, slice(1, None))


    def to_signature(object_path: catalog.Path,
                     module_path: catalog.Path) -> Optional[Base]:
        try:
            object_nodes = arboretum.to_nodes(object_path,
                                              module_path)
        except KeyError:
            return None
        else:
            try:
                signatures = list(to_signatures(object_nodes))
            except AttributeError:
                return None
            else:
                return flatten_signatures(signatures)


    def flatten_signatures(signatures: Iterable[Base]) -> Base:
        signatures = list(signatures)
        try:
            signature, = signatures
        except ValueError:
            return Overloaded(*signatures)
        else:
            return signature


    @slice_parameters.register(Overloaded)
    def slice_overloaded_parameters(signature: Overloaded,
                                    slice_: slice) -> Base:
        return Overloaded(*map(partial(slice_parameters,
                                       slice_=slice_),
                               signature.signatures))


    def from_ast(object_: ast3.arguments) -> Base:
        parameters = to_parameters(object_)
        return Plain(*parameters)


    to_signatures = mapper(compose(from_ast, attrgetter('args')))


    def to_positional_parameters(signature_ast: ast3.arguments
                                 ) -> Iterable[Parameter]:
        # double-reversing since parameters with default arguments go last
        parameters_with_defaults_ast = zip_longest(reverse(signature_ast.args),
                                                   signature_ast.defaults)
        parameters_with_defaults_ast = reverse(parameters_with_defaults_ast)
        parameter_factory = partial(to_parameter,
                                    kind=Parameter.Kind.POSITIONAL_ONLY)
        parameters_factory = mapper(pack(parameter_factory))
        yield from parameters_factory(parameters_with_defaults_ast)


    def to_keyword_parameters(signature_ast: ast3.arguments
                              ) -> Iterable[Parameter]:
        parameters_with_defaults_ast = zip(signature_ast.kwonlyargs,
                                           signature_ast.kw_defaults)
        parameter_factory = partial(to_parameter,
                                    kind=Parameter.Kind.KEYWORD_ONLY)
        parameters_factory = mapper(pack(parameter_factory))
        yield from parameters_factory(parameters_with_defaults_ast)


    def to_variadic_positional_parameter(signature_ast: ast3.arguments
                                         ) -> Optional[Parameter]:
        parameter_ast = signature_ast.vararg
        if parameter_ast is None:
            return None
        return Parameter(name=parameter_ast.arg,
                         kind=Parameter.Kind.VARIADIC_POSITIONAL,
                         has_default=False)


    def to_variadic_keyword_parameter(signature_ast: ast3.arguments
                                      ) -> Optional[Parameter]:
        parameter_ast = signature_ast.kwarg
        if parameter_ast is None:
            return None
        return Parameter(name=parameter_ast.arg,
                         kind=Parameter.Kind.VARIADIC_KEYWORD,
                         has_default=False)


    to_parameters = compose(
            sifter(),
            flatten,
            cleave([to_positional_parameters,
                    compose(expand, to_variadic_positional_parameter),
                    to_keyword_parameters,
                    compose(expand, to_variadic_keyword_parameter)]))


    def to_parameter(parameter_ast: ast3.arg,
                     default_ast: Optional[ast3.expr],
                     *,
                     kind: Parameter.Kind) -> Parameter:
        return Parameter(name=parameter_ast.arg,
                         kind=kind,
                         has_default=default_ast is not None)

from_callable = right.folder(factory.register, from_callable)([
    BuiltinFunctionType,
    BuiltinMethodType,
    FunctionType,
    MethodType,
    MethodDescriptorType,
    WrapperDescriptorType])
from_class_cache = {
    product: Plain(Parameter(name='iterables',
                             kind=Parameter.Kind.VARIADIC_POSITIONAL,
                             has_default=False),
                   Parameter(name='repeat',
                             kind=Parameter.Kind.KEYWORD_ONLY,
                             has_default=True)),
}
from_class = factory.register(type)(cached_map(from_class_cache)(from_class))



import inspect
from functools import (partial,
                       singledispatch)
from itertools import (chain,
                       starmap)
from typing import (Callable,
                    Dict,
                    Tuple,
                    Union)

from . import signatures
from .hints import (Domain,
                    Range)


class Curry:
    def __init__(self,
                 callable_: Callable[..., Range],
                 signature: signatures.Base,
                 *args: Domain,
                 **kwargs: Domain) -> None:
        self.callable_ = callable_
        self.signature = signature
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args: Domain, **kwargs: Domain
                 ) -> Union['Curry', Range]:
        args = self.args + args
        kwargs = {**self.kwargs, **kwargs}
        try:
            return self.callable_(*args, **kwargs)
        except TypeError:
            if not self.signature.has_unset_parameters(*args, **kwargs):
                raise
            return type(self)(self.callable_, self.signature, *args, **kwargs)

    def __eq__(self, other: 'Curry') -> bool:
        if not isinstance(other, Curry):
            return NotImplemented
        return (self.callable_ is other.callable_
                and self.signature == other.signature
                and self.args == other.args
                and self.kwargs == other.kwargs)

    def __repr__(self) -> str:
        result = ('<curried {callable_}'
                  .format(callable_=self.callable_))
        arguments_strings = list(chain(map(str, self.args),
                                       starmap('{}={}'.format,
                                               self.kwargs.items())))
        if arguments_strings:
            result += (' with partially applied {arguments}'
                       .format(arguments=', '.join(arguments_strings)))
        result += '>'
        return result


@singledispatch
def unwrap(object_: Callable[..., Range]
           ) -> Tuple[Callable, Tuple[Domain, ...], Dict[str, Domain]]:
    return inspect.unwrap(object_), (), {}


@unwrap.register(partial)
def unwrap_partial(object_: partial
                   ) -> Tuple[Callable, Tuple[Domain, ...], Dict[str, Domain]]:
    return object_.func, object_.args, object_.keywords


def curry(callable_: Callable[..., Range]) -> Curry:
    """
    Returns curried version of given callable.
    """
    callable_, args, kwargs = unwrap(callable_)
    signature = signatures.factory(callable_)
    return Curry(callable_, signature, *args, **kwargs)



import importlib.abc
import importlib.machinery
import sys
import types
from functools import singledispatch
from types import ModuleType
from typing import Any

from lz.functional import compose
from lz.iterating import (flatten,
                          mapper,
                          sifter)
from . import (catalog,
               dictionaries)
from .hints import Namespace


@singledispatch
def factory(object_: Any) -> Namespace:
    return object_


replacing_modules_names = {
    '_importlib_modulespec': [types.__name__,
                              importlib.abc.__name__,
                              importlib.machinery.__name__]}
if sys.platform == 'win32':
    import nt

    replacing_modules_names['posix'] = [nt.__name__]
to_replacing_modules_names = compose(flatten,
                                     sifter(),
                                     mapper(replacing_modules_names.get))


@factory.register(catalog.Path)
def from_module_path(object_: catalog.Path) -> Namespace:
    modules_names = list(to_replacing_modules_names(object_.parts))
    if modules_names:
        return dictionaries.merge(map(compose(factory,
                                              importlib.import_module),
                                      modules_names))
    return factory(importlib.import_module(str(object_)))


@factory.register(ModuleType)
def from_module(object_: ModuleType) -> Namespace:
    return dict(vars(object_))



import functools
import itertools
from typing import (Any,
                    Callable,
                    Iterable)

from .hints import (Domain,
                    Intermediate,
                    Map,
                    Range)


def identity(argument: Domain) -> Domain:
    """
    Returns object itself.
    """
    return argument


def compose(last_function: Map[Any, Range],
            *front_functions: Callable[..., Any]) -> Callable[..., Range]:
    """
    Returns functions composition.
    """

    def binary_compose(left_function: Map[Intermediate, Range],
                       right_function: Callable[..., Intermediate]
                       ) -> Callable[..., Range]:
        def composition(*args, **kwargs):
            return left_function(right_function(*args, **kwargs))

        return composition

    functions = (last_function,) + front_functions
    return functools.reduce(binary_compose, functions)


def combine(maps: Iterable[Map]) -> Map[Iterable[Domain], Iterable[Range]]:
    """
    Returns function that applies each map to corresponding argument.
    """

    def combined(arguments: Iterable[Domain]) -> Iterable[Range]:
        yield from (map_(argument)
                    for map_, argument in zip(maps, arguments))

    return combined


def pack(function: Callable[..., Range]) -> Map[Iterable[Domain], Range]:
    """
    Creates function that works with single iterable parameter
    by unpacking elements to original function.
    """

    @functools.wraps(function)
    def packed(iterable: Iterable[Domain]) -> Range:
        return function(*iterable)

    return packed


def to_constant(object_: Domain) -> Callable[..., Domain]:
    """
    Returns function that returns given object.
    """

    def constant(*_: Domain, **__: Domain) -> Domain:
        return object_

    return constant


def flip(function: Callable[..., Range]) -> Callable[..., Range]:
    """
    Returns function with positional arguments flipped.
    """

    @functools.wraps(function)
    def flipped(*args, **kwargs) -> Range:
        return function(*reversed(args), **kwargs)

    return flipped


def cleave(functions: Iterable[Callable[..., Range]]
           ) -> Callable[..., Iterable[Range]]:
    """
    Returns function that separately applies
    given functions to the same arguments.
    """

    def cleft(*args, **kwargs) -> Range:
        yield from (function(*args, **kwargs)
                    for function in functions)

    return cleft
